<?php

namespace Cronard;

use Closure;

class Cronard {

    const CRON_DATE  = [ 'min' => 'i'   , 'hour' => 'G'   , 'day' => 'j'   , 'month' => 'n'   , 'week' => 'w'   ];

    const CRON_RANGE = [ 'min' => '0-59', 'hour' => '0-23', 'day' => '1-31', 'month' => '1-12', 'week' => '0-6' ];

    protected Closure $callback;

    protected array $tab = [
        'min'   => "*",
        'hour'  => "*",
        'day'   => "*",
        'month' => "*",
        'week'  => "*",
    ];

    public function __construct(? string $tab = null) 
    {
        if ( $tab ) {
            $this->setTab($tab);
        }
    }

    public function setTab(string $tab) : self 
    {
        list($min, $hour, $day, $month, $week) = explode(' ', $tab);

        $this->tab = [
            'min'   => $min,
            'hour'  => $hour,
            'day'   => $day,
            'month' => $month,
            'week'  => $week,
        ];

        return $this;
    }

    public function reset()
    {
        $this->tab = [
            'min'   => "*",
            'hour'  => "*",
            'day'   => "*",
            'month' => "*",
            'week'  => "*",
        ];

        unset($this->callback);
    }

    public function run(array $arguments = [], ? Closure $callback = null)
    {
        if ( $this->validateCron($this->tab) ) {
            if ( null !== ( $callback = $callback ?? $this->callback ?? null ) ) {
                $retval = $callback->call($this, ...$arguments);
            }
            else {
                throw new \InvalidArgumentException("A proper callback must be provided to run the task, be it in the function call directly or at class-level.");
            }
        }

        $this->reset();

        return $retval ?? false;
    }

    public function validateCron($cron) : bool 
    {
        $now = new \DateTime();
        
        is_string($cron) && ( $cron = array_combine(array_keys(static::CRON_DATE), array_slice(explode(' ', $cron, 6), 0, 5)) );

        foreach($cron as $key => $value) {
            $item_range = $this->getRangeValue($key, $value);

            if ( ! in_array( (int) $now->format(static::CRON_DATE[$key]), $item_range, false) ){
                return false;
            }
        }

        return true;
    }

    public function callback($set = null) {
        return $set !== null ? $this->callback = $set : $this->callback;
    }

    public function min($value) : self 
    {
        $this->tab['min'] = $value;
        
        return $this;
    }

    public function hour($value) : self 
    {
        $this->tab['hour'] = $value;
        
        return $this;
    }

    public function day($value) : self 
    {
        $this->tab['day'] = $value;
        
        return $this;
    }

    public function week($value) : self 
    {
        $this->tab['week'] = $value;
        
        return $this;
    }

    public function month($value) : self 
    {
        $this->tab['month'] = $value;
        
        return $this;
    }

    public function everyMinute($value = "*", $range = "*") : self 
    {
        return $this->everyMin($value, $range);
    }

    /**
     * Running given function every x minute
     * @param mixed $value
     * @param mixed $range
     * @return $this
     */
    public function everyMin($value = "*", $range = "*") : self 
    {
        $this->tab['min'] = "$range/$value";
        
        return $this;
    }

    public function everyEvenMin() : self 
    {
        return $this->everyMin(2);
    }

    public function everyOddMin() : self 
    {
        return $this->everyMin("2", "1-59");
    }

    /**
     * Running given function every x hour
     * @param mixed $value
     * @param mixed $range
     * @return $this
     */
    public function everyHour($value = "*", $range = "*") : self 
    {
        $this->tab['hour'] = "$range/$value";
        
        return $this;
    }

    public function everyEvenHour() : self 
    {
        return $this->everyHour(2);
    }

    public function everyOddHour() : self 
    {
        return $this->everyHour(2, "1-23");
    }

    /**
     * Running given function every x day
     * @param mixed $value
     * @param mixed $range
     * @return $this
     */
    public function everyDay($value = "*", $range = "*") : self 
    {
        $this->tab['day'] = "$range/$value";
        
        return $this;
    }

    public function everyEvenDay() : self 
    {
        return $this->everyDay(2);
    }

    public function everyOddDay() : self 
    {
        return $this->everyDay(2, "1-31");
    }

    /**
     * Running given function every x month
     * @param mixed $value
     * @param mixed $range
     * @return $this
     */
    public function everyMonth($value = "*", $range = "*") : self 
    {
        $this->tab['day'] = "$range/$value";
        
        return $this;
    }

    public function everyEvenMonth() : self 
    {
        return $this->everyMonth(2);
    }

    public function everyOddMonth() : self 
    {
        return $this->everyMonth(2, "1-11");
    }

    /* Filter by days of the week */
    public function sunday() : self 
    {
        $this->tab['week'][] = 0;
        
        return $this;
    }

    public function monday() : self 
    {
        $this->tab['week'][] = 1;
        
        return $this;
    }

    public function tuesday() : self 
    {
        $this->tab['week'][] = 2;
        
        return $this;
    }

    public function wednesday() : self 
    {
        $this->tab['week'][] = 3;
        
        return $this;
    }

    public function thursday() : self 
    {
        $this->tab['week'][] = 4;
        
        return $this;
    }

    public function friday() : self 
    {
        $this->tab['week'][] = 5;
        
        return $this;
    }

    public function saturday() : self 
    {
        $this->tab['week'][] = 6;
        
        return $this;
    }

    public function weekend() : self 
    {
        $this->tab['week'] = array_merge($this->tab['week'], [0, 6]);
        
        return $this;
    }

    public function weekday() : self 
    {
        $this->tab['week'] = array_merge($this->tab['week'], [1, 2, 3, 4, 5]);
        
        return $this;
    }

    /**
     * Evaluate given range from CRON value
     */
    protected function getRangeValue($key, $range) : array
    {
        # Allows every values
        if ( $range === "*" ) {
            list($min, $max) = explode('-', static::CRON_RANGE[$key]);
            
            return range($min, $max);
        }
        # Already an array
        elseif ( is_array($range) ) {
            return array_unique($range);
        }
        # Range of 1-59/2
        elseif( strpos($range , '/') !== false) {
            list($inner_range , $step) = explode('/' , $range);
            list($from, $to) = explode('-' , $inner_range === '*' ? static::CRON_RANGE[$key] : $inner_range);

            return range($from, $to, $step ?: 1);
        }
        # Normal patterns
        else {
            $retval = [];
            
            foreach(explode(',' , $range) as $item) {
                if( strpos($item, '-') !== false ) {
                    list($min , $max) = explode('-' , $item);
                    $retval[] = range($min, $max);
                }
                else {
                    $retval[] = $item;
                }
            }

            return $retval;
        }
    }

}
